@@ -3,6 +3,8 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
+#include <asm/uaccess.h>
+
 #include <linux/a.out.h>
 #include <linux/malloc.h>
 #include <linux/sched.h>
@@ -11,9 +13,6 @@
 #include <asm/pgtable.h>
 
 
-#if LINUX_VERSION_CODE >= 66375
-
-
 unsigned long *
 create_ibcs_tables(char *p, struct linux_binprm *bprm, int ibcs)
 {
@@ -28,24 +27,25 @@ create_ibcs_tables(char *p, struct linux_binprm *bprm, int ibcs)
 	sp -= argc+1;
 	argv = sp;
 	if (!ibcs) {
-		put_user(envp,--sp);
-		put_user(argv,--sp);
+		sp--;
+		put_user(envp, sp);
+		sp--;
+		put_user(argv, sp);
 	}
-	put_user(argc,--sp);
+	sp--;
+	put_user(argc, sp);
 	current->mm->arg_start = (unsigned long) p;
 	while (argc-->0) {
-		put_user(p,argv++);
-		while (get_user(p++)) /* nothing */ ;
+		put_user(p, argv); argv++;
+		p += strlen_user(p);
 	}
 	put_user(NULL,argv);
 	current->mm->arg_end = current->mm->env_start = (unsigned long) p;
 	while (envc-->0) {
-		put_user(p,envp++);
-		while (get_user(p++)) /* nothing */ ;
+		put_user(p, envp); envp++;
+		p += strlen_user(p);
 	}
 	put_user(NULL,envp);
 	current->mm->env_end = (unsigned long) p;
 	return sp;
 }
-
-#endif
