@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 1994  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: binfmt_xout.c,v 1.32 1997/05/10 14:45:53 jaggy Exp $
+ * $Id: binfmt_xout.c,v 1.46 1998/08/19 19:44:10 jaggy Exp $
  * $Source: /u/CVS/ibcs/iBCSemul/binfmt_xout.c,v $
  *
  * This file is based upon code written by Al Longyear for the COFF file
@@ -15,10 +15,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/segment.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/fs.h>
 #include <linux/sched.h>
@@ -33,7 +30,8 @@
 #include <linux/ptrace.h>
 #include <linux/malloc.h>
 #include <linux/personality.h>
-#include <linux/ldt.h>
+#include <linux/file.h>
+#include <asm/ldt.h>
 
 #include <ibcs/ibcs.h>
 #include <ibcs/xout.h>
@@ -80,15 +78,9 @@ static int load_object(struct linux_binprm *bprm,
 
 
 
-#ifdef INIT_MM
 struct linux_binfmt xout_format = {
-	NULL, &mod_use_count_, load_xout_binary, load_xout_library, NULL
+	NULL, &__this_module, load_xout_binary, load_xout_library, NULL
 };
-#else
-struct linux_binfmt xout_format = {
-	load_xout_binary, load_xout_library
-};
-#endif
 
 
 #if defined(XOUT_DEBUG) && defined(XOUT_SEGMENTS)
@@ -104,7 +96,7 @@ print_desc(int which)
 	int type, dpl, i;
 
 	if (which) {
-		lp = (unsigned long *)current->ldt;
+		lp = (unsigned long *)((struct desc_struct*)(current->mm->segments));
 		count = LDT_ENTRIES;
 		printk(KERN_DEBUG "XOUT: LDT\n");
 	} else {
@@ -196,8 +188,10 @@ clear_memory(unsigned long addr, unsigned long size)
 			" length 0x%08lx\n",
 			addr, size);
 #endif
-	while (size-- != 0)
-		put_fs_byte (0, addr++);
+	while (size-- != 0) {
+		put_user(0, (char *)addr);
+		addr++;
+	}
 }
 
 
@@ -213,9 +207,6 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 	struct xseg *seglist;
 	int nsegs, ntext, ndata;
 	int not_pageable;
-#ifndef STACK_TOP
-	int old_fs;
-#endif
 	int i;
 	int status;
 	int fd = -1;
@@ -302,16 +293,8 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 #endif
 		return -ENOEXEC;
 	}
-#ifdef STACK_TOP
-	status = read_exec(bprm->inode, xext->xe_segpos,
+	status = read_exec(bprm->dentry, xext->xe_segpos,
 			(char *)seglist, xext->xe_segsize, 1);
-#else
-	old_fs = get_fs ();
-	set_fs(get_ds());
-	status = read_exec(bprm->inode, xext->xe_segpos,
-			(char *)seglist, xext->xe_segsize);
-	set_fs (old_fs);
-#endif
 	if (status < 0) {
 #ifdef XOUT_DEBUG
 		if (ibcs_trace & TRACE_XOUT_LD)
@@ -326,7 +309,8 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 	/* Check we have at least one text segment and
 	 * determine whether or not it is pageable.
 	 */
-	if (!bprm->inode->i_op || !bprm->inode->i_op->default_file_ops->mmap) {
+	if (!bprm->dentry->d_inode->i_op
+	|| !bprm->dentry->d_inode->i_op->default_file_ops->mmap) {
 #ifdef XOUT_DEBUG
 		if (ibcs_trace & TRACE_XOUT_LD)
 			printk(KERN_DEBUG "XOUT: filesystem does not support"
@@ -357,12 +341,12 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 	/*
 	 *  Fetch a file pointer to the executable.
 	 */
-	fd = open_inode (bprm->inode, O_RDONLY);
+	fd = open_dentry(bprm->dentry, O_RDONLY);
 	if (fd < 0) {
 		kfree(seglist);
 		return fd;
 	} else {
-		fp = current->FD[fd];
+		fp = fcheck(fd);
 	}
 
 	/*
@@ -385,8 +369,8 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 /*
  *  Define the initial locations for the various items in the new process
  */
-		current->MM(mmap)        = NULL;
-		current->MM(rss)         = 0;
+		current->mm->mmap        = NULL;
+		current->mm->rss         = 0;
 /*
  *  Construct the parameter and environment string table entries.
  */
@@ -419,44 +403,32 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 		/* These will be set up later once we've seen the
 		 * segments that make the program up.
 		 */
-		current->MM(start_code)  =
-		current->MM(end_code)    =
-		current->MM(end_data)    =
-		current->MM(start_brk)   =
-		current->MM(brk)         = 0;
+		current->mm->start_code  =
+		current->mm->end_code    =
+		current->mm->end_data    =
+		current->mm->start_brk   =
+		current->mm->brk         = 0;
 #else
-		current->MM(start_code)  = 0;
-		current->MM(end_code)    = xexec->x_text;
-		current->MM(end_data)    = xexec->x_text + xexec->x_data;
-		current->MM(start_brk)   =
-		current->MM(brk)         = xexec->x_text + xexec->x_data + xexec->x_bss;
+		current->mm->start_code  = 0;
+		current->mm->end_code    = xexec->x_text;
+		current->mm->end_data    = xexec->x_text + xexec->x_data;
+		current->mm->start_brk   =
+		current->mm->brk         = xexec->x_text + xexec->x_data + xexec->x_bss;
 #endif
-#ifdef INIT_MM
-		current->suid = current->euid = current->fsuid = bprm->e_uid;
-		current->sgid = current->egid = current->fsgid = bprm->e_gid;
-#else
-		current->suid = current->euid = bprm->e_uid;
-		current->sgid = current->egid = bprm->e_gid;
-#endif
-#ifdef PF_FORKNOEXEC
+		compute_creds(bprm);
 		current->flags &= ~PF_FORKNOEXEC;
-#endif
-#ifndef VM_STACK_FLAGS
-		current->executable  = bprm->inode; /* Store inode for file  */
-		++bprm->inode->i_count;             /* Count the open inode  */
-#endif
 #ifdef XOUT_SEGMENTS
 		/* The code selector is advertised in the header. */
 		if ((xexec->x_cpu & XC_CPU) != XC_386) {
 			regs->ebx = regs->ecx = xext->xe_eseg;
 			regs->eax = xexec->x_entry;
 		} else {
-			regs->cs = regs->ds = regs->es = regs->ss = xext->xe_eseg;
+			regs->xcs = regs->xds = regs->xes = regs->xss = xext->xe_eseg;
 			regs->eip = xexec->x_entry;
 		}
 #else
-		regs->cs = USER_CS;
-		regs->ds = regs->es = regs->ss = USER_DS;
+		regs->xcs = __USER_CS;
+		regs->xds = regs->xes = regs->xss = __USER_DS;
 		regs->eip = xexec->x_entry;
 #endif
 #ifdef XOUT_DEBUG
@@ -465,25 +437,23 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 				xext->xe_eseg, xexec->x_entry);
 #endif
 		regs->esp            =
-		current->MM(start_stack) = bprm->p;
+		current->mm->start_stack = bprm->p;
 
 #ifdef PER_XENIX
 		current->personality = PER_XENIX | PER_SHORT_INODE;
 #else
 		current->personality = PER_SVR3 | PER_SHORT_INODE;
 #endif
-#ifdef INIT_MM
-		if (current->exec_domain && current->exec_domain->use_count)
-			(*current->exec_domain->use_count)--;
-		if (current->binfmt && current->binfmt->use_count)
-			(*current->binfmt->use_count)--;
+		if (current->exec_domain && current->exec_domain->module)
+			__MOD_DEC_USE_COUNT(current->exec_domain->module);
+		if (current->binfmt && current->binfmt->module)
+			__MOD_DEC_USE_COUNT(current->binfmt->module);
 		current->exec_domain = lookup_exec_domain(current->personality);
 		current->binfmt = &xout_format;
-		if (current->exec_domain && current->exec_domain->use_count)
-			(*current->exec_domain->use_count)++;
-		if (current->binfmt && current->binfmt->use_count)
-			(*current->binfmt->use_count)++;
-#endif
+		if (current->exec_domain && current->exec_domain->module)
+			__MOD_INC_USE_COUNT(current->exec_domain->module);
+		if (current->binfmt && current->binfmt->module)
+			__MOD_INC_USE_COUNT(current->binfmt->module);
 	}
 
 	/* Base address for mapping 16bit segments. This should lie above
@@ -497,7 +467,7 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 	 * headers for later reference by the emulator.
 	 */
 	if ((xexec->x_cpu & XC_CPU) != XC_386) {
-		int old_fs = get_fs();
+		mm_segment_t old_fs = get_fs();
 
     		set_fs (get_ds ());
     		status = SYS(uselib)("/usr/lib/x286emul");
@@ -509,8 +479,8 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 			MAP_FIXED|MAP_PRIVATE,
   			0);
 		if (status >= 0) {
-			memcpy_tofs((char *)addr, xexec, sizeof(struct xexec));
-			memcpy_tofs((char *)addr+sizeof(struct xexec), xext, sizeof(struct xext));
+			copy_to_user((char *)addr, xexec, sizeof(struct xexec));
+			copy_to_user((char *)addr+sizeof(struct xexec), xext, sizeof(struct xext));
 			addr = PAGE_ALIGN(addr+sizeof(struct xexec)+sizeof(struct xext));
 		}
 	}
@@ -543,7 +513,7 @@ load_object (struct linux_binprm * bprm, struct pt_regs *regs, int lib_ok)
 			struct modify_ldt_ldt_s ldt_info;
 			struct desc_struct def_ldt;
 			int l;
-			int old_fs = get_fs();
+			mm_segment_t old_fs = get_fs();
 
 seg_again:
 			l = 0;
@@ -577,11 +547,11 @@ seg_again:
 			 */
 			if (s->xs_type == XS_TTEXT) {
 #if 0 /* impossible - start_code is 0 */
-				if (s->xs_rbase < current->MM(start_code))
-					current->MM(start_code) = s->xs_rbase;
+				if (s->xs_rbase < current->mm->start_code)
+					current->mm->start_code = s->xs_rbase;
 #endif
-				if (s->xs_rbase+s->xs_psize > current->MM(end_code))
-					current->MM(end_code) = s->xs_rbase+s->xs_psize;
+				if (s->xs_rbase+s->xs_psize > current->mm->end_code)
+					current->mm->end_code = s->xs_rbase+s->xs_psize;
 			} else if (s->xs_type == XS_TDATA) {
 #ifdef XOUT_SEGMENTS
 				/* If it is the first data segment note that
@@ -597,17 +567,17 @@ seg_again:
 						addr = PAGE_ALIGN(ldt_info.base_addr + s->xs_vsize);
 					}
 				} else {
-					if (regs->ds == regs->cs)
-						regs->ds = regs->es
-						= regs->ss = s->xs_seg;
+					if (regs->xds == regs->xcs)
+						regs->xds = regs->xes
+						= regs->xss = s->xs_seg;
 				}
 #endif
-				if (s->xs_rbase+s->xs_psize > current->MM(end_data))
-					current->MM(end_data) = s->xs_rbase+s->xs_psize;
+				if (s->xs_rbase+s->xs_psize > current->mm->end_data)
+					current->mm->end_data = s->xs_rbase+s->xs_psize;
 			}
-			if (s->xs_rbase+s->xs_vsize > current->MM(brk))
-				current->MM(start_brk) =
-				current->MM(brk) = PAGE_ALIGN(s->xs_rbase+s->xs_vsize);
+			if (s->xs_rbase+s->xs_vsize > current->mm->brk)
+				current->mm->start_brk =
+				current->mm->brk = PAGE_ALIGN(s->xs_rbase+s->xs_vsize);
 #ifdef XOUT_SEGMENTS
 			if (ldt_info.seg_32bit) {
 				ldt_info.limit = (TASK_SIZE-1) >> 12;
@@ -631,7 +601,7 @@ seg_again:
 			 * since we need to copy the lcall7 call gate.
 			 */
 			set_fs(get_ds());
-			if (!current->ldt) {
+			if (!current->mm->segments) {
 				SYS(modify_ldt)(0, &def_ldt, sizeof(def_ldt));
 				l = 1;
 			}
@@ -652,8 +622,8 @@ seg_again:
 			set_fs(old_fs);
 			if (l == 1) {
 				l = 0;
-				current->ldt[0].a = def_ldt.a;
-				current->ldt[0].b = def_ldt.b;
+				((struct desc_struct*)(current->mm->segments))[0].a = def_ldt.a;
+				((struct desc_struct*)(current->mm->segments))[0].b = def_ldt.b;
 			}
 			if (status < 0) {
 				printk(KERN_INFO "XOUT: modify_ldt returned %d\n", status);
@@ -686,15 +656,7 @@ seg_again:
 							: PROT_READ|PROT_WRITE|PROT_EXEC,
 						(s->xs_attr & XS_APURE)
   							? MAP_FIXED|MAP_SHARED
-#ifdef MAP_EXECUTABLE
 							: MAP_FIXED|MAP_PRIVATE | MAP_DENYWRITE | MAP_EXECUTABLE,
-#else
-#  ifdef MAP_DENYWRITE
-							: MAP_FIXED|MAP_PRIVATE | MAP_DENYWRITE,
-#  else
-							: MAP_FIXED|MAP_PRIVATE,
-#  endif
-#endif
   						s->xs_filpos);
 
 					/* Map uninitialised data. */
@@ -718,12 +680,12 @@ seg_again:
 				}
 			}
 #ifdef XOUT_SEGMENTS
-			if (status >= 0 && !ntext && s->xs_seg == 0x47) {
+			if (status >= 0 && !ntext && ndata == 1 && s->xs_seg >= 0x47) {
 				/* Uh oh, impure binary. Mirror this data
 				 * segment to the text segment.
 				 */
-				mirror_addr = s->xs_rbase;
-				s->xs_seg = 0x3f;
+ 				addr = mirror_addr = s->xs_rbase;
+ 				s->xs_seg = xext->xe_eseg;
 				s->xs_type = XS_TTEXT;
 				goto seg_again;
 			}
@@ -732,16 +694,16 @@ seg_again:
 	}
 
 #if 0
-	if (addr > current->MM(brk))
-		current->MM(start_brk) = current->MM(brk) = addr;
+	if (addr > current->mm->brk)
+		current->mm->start_brk = current->mm->brk = addr;
 #endif
 
 #ifdef XOUT_DEBUG
 	if (ibcs_trace & TRACE_XOUT_LD) {
 		printk(KERN_DEBUG "XOUT: start code 0x%08lx, end code 0x%08lx,"
 			" end data 0x%08lx, brk 0x%08lx\n",
-			current->MM(start_code), current->MM(end_code),
-			current->MM(end_data), current->MM(brk));
+			current->mm->start_code, current->mm->end_code,
+			current->mm->end_data, current->mm->brk);
 #ifdef XOUT_SEGMENTS
 		print_desc(1);
 		print_desc(0);
@@ -758,39 +720,29 @@ seg_again:
 					printk(KERN_DEBUG "XOUT: read to 0x%08lx from 0x%08lx, length 0x%08lx\n",
 						seglist[i].xs_rbase, seglist[i].xs_filpos, seglist[i].xs_psize);
 #endif
-#ifdef STACK_TOP
 				if (seglist[i].xs_psize > 0
-				&& (unsigned long)read_exec(bprm->inode, seglist[i].xs_filpos,
+				&& (unsigned long)read_exec(bprm->dentry, seglist[i].xs_filpos,
 					(char *)seglist[i].xs_rbase,
 					seglist[i].xs_psize, 0) != seglist[i].xs_psize) {
-#else
-				if (seglist[i].xs_psize > 0
-				&& (unsigned long)read_exec(bprm->inode, seglist[i].xs_filpos,
-					(char *)seglist[i].xs_rbase,
-					seglist[i].xs_psize) != seglist[i].xs_psize) {
-#endif
-#ifdef XOUT_DEBUG
-					if (ibcs_trace & TRACE_XOUT_LD)
-						printk(KERN_DEBUG "XOUT: short read\n");
-#endif
-					status = -1;
-				} else if (mirror_addr
-				&& seglist[i].xs_seg == 0x3f
-#ifdef STACK_TOP
-				&& (unsigned long)read_exec(bprm->inode, seglist[i].xs_filpos,
-				(char *)mirror_addr,
-				seglist[i].xs_psize, 0) != seglist[i].xs_psize) {
-#else
-				&& (unsigned long)read_exec(bprm->inode, seglist[i].xs_filpos,
-				(char *)mirror_addr,
-				seglist[i].xs_psize) != seglist[i].xs_psize) {
-#endif
 #ifdef XOUT_DEBUG
 					if (ibcs_trace & TRACE_XOUT_LD)
 						printk(KERN_DEBUG "XOUT: short read\n");
 #endif
 					status = -1;
 				}
+#if 0
+				else if (mirror_addr
+				&& seglist[i].xs_seg == 0x3f
+				&& (unsigned long)read_exec(bprm->dentry, seglist[i].xs_filpos,
+				(char *)mirror_addr,
+				seglist[i].xs_psize, 0) != seglist[i].xs_psize) {
+#ifdef XOUT_DEBUG
+					if (ibcs_trace & TRACE_XOUT_LD)
+						printk(KERN_DEBUG "XOUT: short read\n");
+#endif
+					status = -1;
+				}
+#endif
 			}
 		}
 	}
@@ -833,10 +785,10 @@ seg_again:
 	if ((xexec->x_cpu & XC_CPU) != XC_386) {
 #if 0
 		regs->eax = regs->eip;
-		regs->ebx = regs->cs;
-		regs->ecx = regs->ds;
-		regs->cs = USER_CS;
-		regs->ds = regs->es = regs->ss = regs->fs = USER_DS;
+		regs->ebx = regs->xcs;
+		regs->ecx = regs->xds;
+		regs->xcs = __USER_CS;
+		regs->xds = regs->xes = regs->xss = __USER_DS;
 #endif
 		regs->eip = 0x1020;
 #ifdef XOUT_DEBUG
@@ -848,14 +800,14 @@ seg_again:
 #endif
 #ifdef XOUT_DEBUG
 		if (ibcs_trace & TRACE_XOUT_DB)
-			while (!current->signal) schedule();
+			while (!signal_pending(current)) schedule();
 #endif
 		return (status < 0 ? status : regs->eax);
 	}
 
 #ifdef XOUT_DEBUG
 	if (ibcs_trace & TRACE_XOUT_DB)
-		while (!current->signal) schedule();
+		while (!signal_pending(current)) schedule();
 #endif
 	/* Xenix 386 programs expect the initial brk value to be in eax
 	 * on start up. Hence if we succeeded we need to pass back
@@ -863,7 +815,7 @@ seg_again:
 	 * ret_from_sys_call assembly will place this in eax before
 	 * resuming (starting) the process.
 	 */
-	return (status < 0 ? status : current->MM(brk));
+	return (status < 0 ? status : current->mm->brk);
 }
 
 
@@ -910,28 +862,19 @@ load_xout_library(int fd)
 	} else {
 		struct file *file;
 		struct pt_regs regs;
-#ifndef STACK_TOP
-		int old_fs = get_fs ();
-#endif
 
 		memset(bprm, '\0', sizeof (struct linux_binprm));
 
-		file           = current->FD[fd];
-		bprm->inode    = file->f_inode;
+		file           = fget(fd);
+		bprm->dentry   = file->f_dentry;
 		bprm->filename = "";
 
-#ifdef STACK_TOP
-		status = read_exec(bprm->inode, 0L,
+		status = read_exec(bprm->dentry, 0L,
 			    bprm->buf, sizeof (bprm->buf), 1);
-#else
-		set_fs(get_ds());
-		status = read_exec(bprm->inode, 0L,
-			    bprm->buf, sizeof (bprm->buf));
-		set_fs(old_fs);
-#endif
 
 		status = load_object(bprm, &regs, 0);
 
+		fput(file);
 		kfree(bprm);
 	}
 
