@@ -13,10 +13,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/segment.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/ptrace.h>
 #include <linux/errno.h>
@@ -75,7 +72,7 @@ static int ibcs_hrtalarm(struct pt_regs * regs) {
   struct itimerval get_buffer;
   struct hrtcmd * hcmd;
   int i, error, cmd, retval, which;
-  int old_fs = get_fs();
+  mm_segment_t old_fs = get_fs();
 
   i = get_syscall_parameter (regs, 2);
   if(i != 1) return -EINVAL;
@@ -86,10 +83,10 @@ static int ibcs_hrtalarm(struct pt_regs * regs) {
   if (error)
     return error;
 
-  cmd =  get_fs_long(((unsigned long *) hcmd));
+  get_user(cmd, ((unsigned long *) hcmd));
 
   /* Now figure out which clock we want to fiddle with */
-  which =  get_fs_long(((unsigned long *) hcmd)+1);
+  get_user(which, ((unsigned long *) hcmd)+1);
 
 #ifdef IBCS_TRACE
   if ((ibcs_trace & TRACE_API) || ibcs_func_p->trace) {
@@ -114,8 +111,8 @@ static int ibcs_hrtalarm(struct pt_regs * regs) {
 
   switch(cmd) {
   case 0xc:
-    if(get_fs_long(((unsigned long *) hcmd)+4) != 1000000) return -EINVAL;
-    memcpy_fromfs(&get_buffer.it_value, ((unsigned long *) hcmd)+2,
+    if(({long r; get_user(r, ((unsigned long *) hcmd)+4); r;}) != 1000000) return -EINVAL;
+    copy_from_user(&get_buffer.it_value, ((unsigned long *) hcmd)+2,
 		   sizeof(struct timeval));
     memset(&get_buffer.it_interval, 0, sizeof(struct timeval));
     set_fs(get_ds());
@@ -136,17 +133,17 @@ static int ibcs_hrtalarm(struct pt_regs * regs) {
 	   get_buffer.it_value.tv_usec);
     }
 #endif
-    put_fs_long(1000000, &hcmd->interval.resolution);
-    memcpy_tofs(((unsigned long *) hcmd)+2, &get_buffer.it_interval,
+    put_user(1000000, &hcmd->interval.resolution);
+    copy_to_user(((unsigned long *) hcmd)+2, &get_buffer.it_interval,
 		   sizeof(get_buffer));
     retval = 1;
     break;
   case 0xf:
-    if(get_fs_long(((unsigned long *) hcmd)+4) != 1000000) return -EINVAL;
-    if(get_fs_long(((unsigned long *) hcmd)+7) != 1000000) return -EINVAL;
-    memcpy_fromfs(&get_buffer.it_value, &hcmd->tod,
+    if(({long r; get_user(r, ((unsigned long *) hcmd)+4); r;}) != 1000000) return -EINVAL;
+    if(({long r; get_user(r, ((unsigned long *) hcmd)+7); r;}) != 1000000) return -EINVAL;
+    copy_from_user(&get_buffer.it_value, &hcmd->tod,
 		   sizeof(struct timeval));
-    memcpy_fromfs(&get_buffer.it_interval, &hcmd->interval,
+    copy_from_user(&get_buffer.it_interval, &hcmd->interval,
 		   sizeof(struct timeval));
     set_fs(get_ds());
     retval = SYS(setitimer)(which, &get_buffer, NULL);
