@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 1994  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: map.c,v 1.11 1997/08/16 14:05:59 jaggy Exp $
+ * $Id: map.c,v 1.13 1998/06/09 21:57:26 jaggy Exp $
  * $Source: /u/CVS/ibcs/iBCSemul/map.c,v $
  */
 #include <linux/config.h>
@@ -29,13 +29,6 @@
 /* Error code maps. */
 #include "maps/errmap.inc"
 
-/* Signal maps - these are pulled in to emulate.c where they are need to
- * initialise the execution domain descriptors with 1.1 kernels.
- */
-#ifndef INIT_MM
-#include "maps/signal.inc"
-#endif
-
 /* Socket subsystem maps. */
 #include "maps/sockaf.inc"		/* Address families/formats */
 #include "maps/sockopt.inc"		/* Socket options */
@@ -57,6 +50,37 @@ map_bitvec(unsigned long vec, long map[])
 }
 
 
+unsigned long
+map_sigvec_from_kernel(sigset_t vec, unsigned long map[])
+{
+	unsigned long newvec;
+	int i;
+
+	newvec = 0;
+	for (i=1; i<=32; i++) {
+		if (sigismember(&vec, i) && map[i] != -1)
+			newvec |= (1 << map[i]);
+	}
+	return newvec;
+}
+
+
+sigset_t
+map_sigvec_to_kernel(unsigned long vec, unsigned long map[])
+{
+	sigset_t newvec;
+	unsigned long m;
+	int i;
+
+	sigemptyset(&newvec);
+	for (m=1,i=1; i<=32; m<<=1,i++) {
+		if ((vec & m) && map[i] != -1)
+			sigaddset(&newvec, map[i]);
+	}
+	return newvec;
+}
+
+
 int
 map_value(struct map_segment *m[], int val, int def) {
 	int pers;
