@@ -11,7 +11,7 @@
  *  Please consider this closely if you plan on changing this mode.
  *  -- Al Longyear
  *
- * $Id: signal.c,v 1.25 1997/02/18 22:05:38 jaggy Exp $
+ * $Id: signal.c,v 1.33 1998/10/15 21:20:46 jaggy Exp $
  * $Source: /u/CVS/ibcs/iBCSemul/signal.c,v $
  */
 
@@ -20,10 +20,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/segment.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/errno.h>
 #include <linux/sched.h>
@@ -38,6 +35,7 @@
 #include <asm/system.h>
 #include <linux/fs.h>
 #include <linux/sys.h>
+#include <linux/signal.h>
 
 #include <ibcs/ibcs.h>
 #include <ibcs/xnx.h>
@@ -77,21 +75,6 @@ typedef void (*pfn) (void);     /* Completion function */
 /* Return a mask that includes SIG only.  */
 #define __sigmask(sig)	(1 << ((sig) - 1))
 
-/*
- *  Defines for the other signal processing routines
- */
-
-#define __sigemptyset(set)	((*(set) = 0L), 0)
-#define __sigfillset(set)       ((*(set) = -1L), 0)
-#define __sigaddset(set, sig)   ((*(set) |= __sigmask (sig)), 0)
-#define __sigdelset(set, sig)   ((*(set) &= ~__sigmask (sig)), 0)
-#define __sigismember(set, sig) ((*(set) & __sigmask (sig)) ? 1 : 0)
-
-#define sigaddset	__sigaddset
-#define sigdelset	__sigdelset
-#define sigismember	__sigismember
-#define sigemptyset	__sigemptyset
-#define sigfillset	__sigfillset
 
 #define TO_KERNEL(save)      \
 	save = get_fs ();    \
@@ -104,33 +87,22 @@ typedef void (*pfn) (void);     /* Completion function */
 
 #define _BLOCKABLE (~(_S(IBCS_SIGKILL) | _S(IBCS_SIGSTOP)))
 
-
-#ifndef INIT_MM
-/*
- * Translate the signal number from Linux to iBCS. (Only used in the
- * 1.0 kernel patches to build the iBCS signal stack frame).
- */
-int ibcs_lmapsig(int sig)
+void
+deactivate_signal(struct task_struct *task, int signum)
 {
-	if ((unsigned int)sig >= NSIGNALS)
-		return -1;
-	return signal_map_from_linux[current->personality & PER_MASK][sig];
+	spin_lock_irq(&task->sigmask_lock);
+	sigdelset(&task->signal, signum);
+	spin_unlock_irq(&task->sigmask_lock);
 }
-#endif /* INIT_MM */
 
 /*
  *  Translate the signal number to the corresponding item for Linux.
  */
 static int ibcs_mapsig(int sig)
 {
-
 	if ((unsigned int) sig >= NSIGNALS)
 		return -1;
-#ifdef INIT_MM
 	return current->exec_domain->signal_map[sig];
-#else
-	return signal_map_to_linux[current->personality & PER_MASK][sig];
-#endif
 }
 
 inline int ibcs_signo (struct pt_regs *regs, int *sig)
@@ -159,7 +131,7 @@ void ibcs_sig_handler (struct pt_regs * regs, int sig,
 {
 	struct sigaction act, oact;
 	int	      answer;
-	int	      old_fs;
+	mm_segment_t old_fs;
 
 	sigemptyset (&act.sa_mask);
 	act.sa_restorer = NULL;
@@ -172,7 +144,7 @@ void ibcs_sig_handler (struct pt_regs * regs, int sig,
 		act.sa_flags = 0;
 
 	TO_KERNEL (old_fs);
-	answer = SYS(sigaction) (sig, &act, &oact);
+	answer = SYS(rt_sigaction) (sig, &act, &oact, sizeof(sigset_t));
 	FROM_KERNEL (old_fs);
 
 	if (answer < 0) {
@@ -210,14 +182,12 @@ int ibcs_sigset (struct pt_regs * regs)
 	sigset_t	 newmask, oldmask;
 	__sighandler_t   vec;
 	int	      sig, answer;
-	int	      old_fs;
+	mm_segment_t old_fs;
 
 	if (ibcs_signo (regs, &sig)) {
 		vec = (__sighandler_t) SECOND_PARAM;
 		if (vec != SIG_HOLD) {
-			sigemptyset (&newmask);
-			sigaddset  (&newmask, sig);
-			current->blocked &= ~newmask;
+			deactivate_signal(current, sig);
 			ibcs_sig_handler (regs, sig, vec, 0);
 		} else {
 /*
@@ -227,9 +197,9 @@ int ibcs_sigset (struct pt_regs * regs)
 			sigaddset  (&newmask, sig);
 
 			TO_KERNEL (old_fs);
-			answer = SYS(sigprocmask) (SIG_BLOCK,
-						  &newmask,
-						  &oldmask);
+			answer = SYS(rt_sigprocmask) (SIG_BLOCK,
+						&newmask, &oldmask,
+						sizeof(sigset_t));
 			FROM_KERNEL (old_fs);
 
 			if (answer < 0) {
@@ -249,7 +219,7 @@ void ibcs_sighold (struct pt_regs * regs)
 {
 	sigset_t   newmask, oldmask;
 	int	sig, answer;
-	int	old_fs;
+	mm_segment_t old_fs;
 
 	if (!ibcs_signo (regs, &sig))
 		return;
@@ -258,7 +228,8 @@ void ibcs_sighold (struct pt_regs * regs)
 	sigaddset  (&newmask, sig);
 
 	TO_KERNEL (old_fs);
-	answer = SYS(sigprocmask) (SIG_BLOCK, &newmask, &oldmask);
+	answer = SYS(rt_sigprocmask) (SIG_BLOCK, &newmask, &oldmask,
+				sizeof(sigset_t));
 	FROM_KERNEL (old_fs);
 
 	if (answer < 0) {
@@ -279,7 +250,7 @@ void ibcs_sigrelse (struct pt_regs * regs)
 {
 	sigset_t   newmask, oldmask;
 	int	sig, answer;
-	int	old_fs;
+	mm_segment_t old_fs;
 
 	if (!ibcs_signo (regs, &sig))
 		return;
@@ -288,7 +259,8 @@ void ibcs_sigrelse (struct pt_regs * regs)
 	sigaddset   (&newmask, sig);
 
 	TO_KERNEL (old_fs);
-	answer = SYS(sigprocmask) (SIG_UNBLOCK, &newmask, &oldmask);
+	answer = SYS(rt_sigprocmask) (SIG_UNBLOCK, &newmask, &oldmask,
+				sizeof(sigset_t));
 	FROM_KERNEL (old_fs);
 
 	if (answer < 0) {
@@ -306,7 +278,7 @@ void ibcs_sigignore (struct pt_regs * regs)
 {
 	struct sigaction act, oact;
 	int	      sig, answer;
-	int	      old_fs;
+	mm_segment_t old_fs;
 
 	if (!ibcs_signo (regs, &sig))
 		return;
@@ -318,7 +290,7 @@ void ibcs_sigignore (struct pt_regs * regs)
 	act.sa_flags   = 0;
 
 	TO_KERNEL (old_fs);
-	answer = SYS(sigaction) (sig, &act, &oact);
+	answer = SYS(rt_sigaction) (sig, &act, &oact, sizeof(sigset_t));
 	FROM_KERNEL (old_fs);
 
 	if (answer < 0) {
@@ -338,7 +310,7 @@ void ibcs_sigignore (struct pt_regs * regs)
 
 void ibcs_sigpause (struct pt_regs * regs)
 {
-	sigset_t   newset;
+	old_sigset_t   newset;
 	int	sig, answer;
 
 #ifdef __sparc__
@@ -347,19 +319,18 @@ void ibcs_sigpause (struct pt_regs * regs)
 	if (!ibcs_signo(regs, &sig))
 		return;
 
-	sigfillset(&newset);
-	sigdelset(&newset, sig);
+	newset = ~0UL;
+	newset &= (1UL << (sig-1));
 	answer = SYS(sigsuspend)(0, current->blocked,
 			newset, regs->esi, regs->edi,
 			regs->ebp, regs->eax,
-			regs->ds, regs->es,
-			regs->fs, regs->gs, regs->orig_eax,
-			regs->eip, regs->cs, regs->eflags,
-			regs->esp, regs->ss);
+			regs->xds, regs->xes,
+			regs->orig_eax,
+			regs->eip, regs->xcs, regs->eflags,
+			regs->esp, regs->xss);
 
 	if (answer < 0) {
-		regs->eax = iABI_errors(-answer);
-		regs->eflags |= 1;
+		set_error(regs, iABI_errors(-answer));
 	}
 #endif
 }
@@ -470,179 +441,145 @@ asmlinkage int abi_sigaction(int abi_signum, const struct abi_sigaction * action
 	struct abi_sigaction * oldaction)
 {
 	struct abi_sigaction new_sa, old_sa;
-	int signum;
-	struct sigaction *p;
+	int error, signo;
+	mm_segment_t old_fs;
+	struct sigaction nsa, osa;
 
-	signum = ibcs_mapsig(abi_signum);
-	if (signum<1 || signum>32 || signum==SIGKILL || signum==SIGSTOP)
+	signo = ibcs_mapsig(abi_signum);
+	if (signo == -1)
 		return -EINVAL;
-	p = signum - 1 + current->SIGACTION;
-	if (action) {
-		memcpy_fromfs(&new_sa, action, sizeof(struct abi_sigaction));
-		if (new_sa.sa_flags & ABI_SA_NODEFER)
-			new_sa.sa_mask = 0;
-		else {
-			new_sa.sa_mask |= _S(abi_signum);
-			new_sa.sa_mask &= _BLOCKABLE;
-		}
-		if (TASK_SIZE <= (unsigned long) new_sa.sa_handler)
-			return -EFAULT;
-	}
-	if (oldaction) {
-		if (!verify_area(VERIFY_WRITE,oldaction, sizeof(struct abi_sigaction))) {
-		  old_sa.sa_handler = p->sa_handler;
-#ifdef INIT_MM
-		  old_sa.sa_mask = map_bitvec(p->sa_mask,
-			current->exec_domain->signal_invmap);
-#else
-		  old_sa.sa_mask = map_bitvec(p->sa_mask,
-			signal_map_from_linux[current->personality & PER_MASK]);
-#endif
-		  old_sa.sa_flags = 0;
-		  if(p->sa_flags & SA_STACK)
-			old_sa.sa_flags |= ABI_SA_ONSTACK;
-		  if(p->sa_flags & SA_RESTART)
-			old_sa.sa_flags |= ABI_SA_RESTART;
-		  if(p->sa_flags & SA_NOMASK)
-			old_sa.sa_flags |= ABI_SA_NODEFER;
-		  if(p->sa_flags & SA_ONESHOT)
-			old_sa.sa_flags |= ABI_SA_RESETHAND;
-		  if(p->sa_flags & SA_NOCLDSTOP)
-			old_sa.sa_flags |= ABI_SA_NOCLDSTOP;
-		  memcpy_tofs(oldaction, &old_sa, sizeof(struct abi_sigaction));
-		      };
-	}
-	if (action) {
-	        /* The internal format of the sigaction structure is
-		   different, so we cannot simply copy the structure. */
-		p->sa_restorer = NULL;
-	  	p->sa_handler = new_sa.sa_handler;
-#ifdef INIT_MM
-		  p->sa_mask = map_bitvec(new_sa.sa_mask,
-			current->exec_domain->signal_map);
-#else
-		  p->sa_mask = map_bitvec(new_sa.sa_mask,
-			signal_map_to_linux[current->personality & PER_MASK]);
-#endif
-		p->sa_flags = 0;
-		if(new_sa.sa_flags & ABI_SA_ONSTACK)
-			p->sa_flags |= SA_STACK;
-		if(new_sa.sa_flags & ABI_SA_RESTART)
-			p->sa_flags |= SA_RESTART;
-		if(new_sa.sa_flags & ABI_SA_NODEFER)
-			p->sa_flags |= SA_NOMASK;
-		if(new_sa.sa_flags & ABI_SA_RESETHAND)
-			p->sa_flags |= SA_ONESHOT;
-		if(new_sa.sa_flags & ABI_SA_NOCLDSTOP)
-			p->sa_flags |= SA_NOCLDSTOP;
 
-		/* Check for pending signals. */
-		if (p->sa_handler == SIG_IGN) {
-			if (signum == SIGCHLD)
-				return 0;
-			current->signal &= ~_S(signum);
-			return 0;
-		}
-		if (p->sa_handler == SIG_DFL) {
-			if (signum != SIGCONT
-			&& signum != SIGCHLD
-			&& signum != SIGWINCH)
-				return 0;
-			current->signal &= ~_S(signum);
-			return 0;
-		}	
+	if (oldaction) {
+		error = verify_area(VERIFY_WRITE, oldaction,
+				sizeof(struct abi_sigaction));
+		if (error)
+			return error;
 	}
-	return 0;
+
+	if (action) {
+		error = copy_from_user(&new_sa, action,
+				sizeof(struct abi_sigaction));
+		if (error)
+			return -EFAULT;
+		nsa.sa_restorer = NULL;
+		nsa.sa_handler = new_sa.sa_handler;
+		nsa.sa_mask = map_sigvec_to_kernel(new_sa.sa_mask,
+			current->exec_domain->signal_map);
+		if(new_sa.sa_flags & ABI_SA_ONSTACK)
+			nsa.sa_flags |= SA_ONSTACK;
+		if(new_sa.sa_flags & ABI_SA_RESTART)
+			nsa.sa_flags |= SA_RESTART;
+		if(new_sa.sa_flags & ABI_SA_NODEFER)
+			nsa.sa_flags |= SA_NODEFER;
+		if(new_sa.sa_flags & ABI_SA_RESETHAND)
+			nsa.sa_flags |= SA_RESETHAND;
+		if(new_sa.sa_flags & ABI_SA_NOCLDSTOP)
+			nsa.sa_flags |= SA_NOCLDSTOP;
+		if(new_sa.sa_flags & ABI_SA_NOCLDWAIT)
+			nsa.sa_flags |= SA_NOCLDWAIT;
+	}
+
+	old_fs = get_fs();
+	set_fs(get_ds());
+	error = SYS(rt_sigaction)(signo,
+				action ? &nsa : NULL,
+				oldaction ? &osa : NULL,
+				sizeof(sigset_t));
+	set_fs(old_fs);
+
+	if (!error && oldaction) {
+		old_sa.sa_handler = osa.sa_handler;
+		old_sa.sa_mask = map_sigvec_from_kernel(osa.sa_mask,
+			current->exec_domain->signal_invmap);
+		old_sa.sa_flags = 0;
+		if(osa.sa_flags & SA_ONSTACK)
+			old_sa.sa_flags |= ABI_SA_ONSTACK;
+		if(osa.sa_flags & SA_RESTART)
+			old_sa.sa_flags |= ABI_SA_RESTART;
+		if(osa.sa_flags & SA_NODEFER)
+			old_sa.sa_flags |= ABI_SA_NODEFER;
+		if(osa.sa_flags & SA_RESETHAND)
+			old_sa.sa_flags |= ABI_SA_RESETHAND;
+		if(osa.sa_flags & SA_NOCLDSTOP)
+			old_sa.sa_flags |= ABI_SA_NOCLDSTOP;
+		if(osa.sa_flags & SA_NOCLDWAIT)
+			old_sa.sa_flags |= ABI_SA_NOCLDWAIT;
+		/* This should never fail... */
+		copy_to_user(oldaction, &old_sa, sizeof(struct abi_sigaction));
+	}
+	return error;
 }
 
 asmlinkage int sco_sigaction(int sco_signum, const struct sco_sigaction * action,
 	struct sco_sigaction * oldaction)
 {
 	struct sco_sigaction new_sa, old_sa;
-	int signum;
-	struct sigaction *p;
+	int error, signo;
+	mm_segment_t old_fs;
+	struct sigaction nsa, osa;
 
-	signum = ibcs_mapsig(sco_signum);
-	if (signum<1 || signum>32 || signum==SIGKILL || signum==SIGSTOP)
+	signo = ibcs_mapsig(sco_signum);
+	if (signo == -1)
 		return -EINVAL;
-	p = signum - 1 + current->SIGACTION;
-	if (action) {
-		memcpy_fromfs(&new_sa, action, sizeof(struct sco_sigaction));
-		new_sa.sa_mask |= _S(sco_signum);
-		new_sa.sa_mask &= _BLOCKABLE;
-		if (TASK_SIZE <= (unsigned long) new_sa.sa_handler)
-			return -EFAULT;
-	}
-	if (oldaction) {
-		if (!verify_area(VERIFY_WRITE, oldaction, sizeof(struct sco_sigaction))) {
-		  old_sa.sa_handler = p->sa_handler;
-#ifdef INIT_MM
-		  old_sa.sa_mask = map_bitvec(p->sa_mask,
-			current->exec_domain->signal_invmap);
-#else
-		  old_sa.sa_mask = map_bitvec(p->sa_mask,
-			signal_map_from_linux[current->personality & PER_MASK]);
-#endif
-		  old_sa.sa_flags = 0;
-		  if(p->sa_flags & SA_NOCLDSTOP)
-			old_sa.sa_flags |= SCO_SA_NOCLDSTOP;
-		  memcpy_tofs(oldaction, &old_sa, sizeof(struct sco_sigaction));
-		      };
-	}
-	if (action) {
-	        /* The internal format of the sigaction structure is
-		   different, so we cannot simply copy the structure. */
-		p->sa_restorer = NULL;
-	  	p->sa_handler = new_sa.sa_handler;
-#ifdef INIT_MM
-		  p->sa_mask = map_bitvec(new_sa.sa_mask,
-			current->exec_domain->signal_map);
-#else
-		  p->sa_mask = map_bitvec(new_sa.sa_mask,
-			signal_map_to_linux[current->personality & PER_MASK]);
-#endif
-		p->sa_flags = SA_NOMASK;
-		if (new_sa.sa_flags & SCO_SA_NOCLDSTOP)
-			p->sa_flags |= SA_NOCLDSTOP;
 
-		/* Check for pending signals. */
-		if (p->sa_handler == SIG_IGN) {
-			if (signum == SIGCHLD)
-				return 0;
-			current->signal &= ~_S(signum);
-			return 0;
-		}
-		if (p->sa_handler == SIG_DFL) {
-			if (signum != SIGCONT
-			&& signum != SIGCHLD
-			&& signum != SIGWINCH)
-				return 0;
-			current->signal &= ~_S(signum);
-			return 0;
-		}	
+	if (oldaction) {
+		error = verify_area(VERIFY_WRITE, oldaction,
+				sizeof(struct sco_sigaction));
+		if (error)
+			return error;
 	}
-	return 0;
+
+	if (action) {
+		error = copy_from_user(&new_sa, action,
+				sizeof(struct sco_sigaction));
+		if (error)
+			return -EFAULT;
+		nsa.sa_restorer = NULL;
+		nsa.sa_handler = new_sa.sa_handler;
+		nsa.sa_mask = map_sigvec_to_kernel(new_sa.sa_mask,
+			current->exec_domain->signal_map);
+		nsa.sa_flags = SA_NOMASK;
+		if (new_sa.sa_flags & SCO_SA_NOCLDSTOP)
+			nsa.sa_flags |= SA_NOCLDSTOP;
+	}
+
+	old_fs = get_fs();
+	set_fs(get_ds());
+	error = SYS(rt_sigaction)(signo,
+				action ? &nsa : NULL,
+				oldaction ? &osa : NULL,
+				sizeof(sigset_t));
+	set_fs(old_fs);
+
+	if (!error && oldaction) {
+		old_sa.sa_handler = osa.sa_handler;
+		old_sa.sa_mask = map_sigvec_from_kernel(osa.sa_mask,
+			current->exec_domain->signal_invmap);
+		old_sa.sa_flags = 0;
+		if(osa.sa_flags & SA_NOCLDSTOP)
+			old_sa.sa_flags |= SCO_SA_NOCLDSTOP;
+		/* This should never fail... */
+		copy_to_user(oldaction, &old_sa, sizeof(struct sco_sigaction));
+	}
+	return error;
 }
 
 static short int howcnv[] = {SIG_SETMASK, SIG_BLOCK, SIG_UNBLOCK, SIG_SETMASK};
 
 asmlinkage int
-abi_sigprocmask(int how, sigset_t *abinset, sigset_t *abioset)
+abi_sigprocmask(int how, unsigned long *abinset, unsigned long *abioset)
 {
 	sigset_t new_set, *nset, old_set, *oset;
-	int old_fs, error;
+	unsigned long new_set_abi, old_set_abi;
+	mm_segment_t old_fs;
+	int error;
 
 	nset = oset = NULL;
 
 	if (abinset) {
-		new_set = get_fs_long((unsigned long *)abinset);
-#ifdef INIT_MM
-		new_set = map_bitvec(new_set,
+		get_user(new_set_abi, abinset);
+		new_set = map_sigvec_to_kernel(new_set_abi,
 			current->exec_domain->signal_map);
-#else
-		new_set = map_bitvec(new_set,
-			signal_map_to_linux[current->personality & PER_MASK]);
-#endif
 		nset = &new_set;
 	}
 	if (abioset)
@@ -650,18 +587,13 @@ abi_sigprocmask(int how, sigset_t *abinset, sigset_t *abioset)
 
 	old_fs = get_fs();
 	set_fs(get_ds());
-	error = SYS(sigprocmask)(howcnv[how], nset, oset);
+	error = SYS(rt_sigprocmask)(howcnv[how], nset, oset, sizeof(sigset_t));
 	set_fs(old_fs);
 
 	if (!error && abioset) {
-#ifdef INIT_MM
-		old_set = map_bitvec(old_set,
+		old_set_abi = map_sigvec_from_kernel(old_set,
 			current->exec_domain->signal_invmap);
-#else
-		old_set = map_bitvec(old_set,
-			signal_map_from_linux[current->personality & PER_MASK]);
-#endif
-		put_fs_long(old_set, (unsigned long *)abioset);
+		put_user(old_set_abi, abioset);
 	}
 
 	return error;
@@ -670,28 +602,24 @@ abi_sigprocmask(int how, sigset_t *abinset, sigset_t *abioset)
 #ifndef __sparc__
 int abi_sigsuspend(struct pt_regs * regs)
 {
-	sigset_t * set;
-	unsigned long newset, oldset;
+	unsigned long * set;
+	unsigned long oldset;
+	old_sigset_t newset;
 	int error;
 
 #if defined(EMU_BSD) && defined(PER_BSD)
-	if (current->personality == PER_BSD) {
+	if (personality(PER_BSD)) {
 		oldset = get_syscall_parameter (regs, 0);
 	} else
 #endif
 	{
-		set = (sigset_t *)get_syscall_parameter (regs, 0);
-		if ((error = verify_area(VERIFY_READ, set, sizeof(sigset_t))))
+		set = (unsigned long *)get_syscall_parameter (regs, 0);
+		error = get_user(oldset, set);
+		if (error)
 			return error;
-		oldset = get_fs_long ((unsigned long *) set);
 	}
-#ifdef INIT_MM
 	newset = map_bitvec(oldset,
 		current->exec_domain->signal_map);
-#else
-	newset = map_bitvec(oldset,
-		signal_map_to_linux[current->personality & PER_MASK]);
-#endif
 
 #ifdef IBCS_TRACE
 	if ((ibcs_trace & TRACE_SIGNAL) || ibcs_func_p->trace)
@@ -707,9 +635,9 @@ int abi_sigsuspend(struct pt_regs * regs)
 	return SYS(sigsuspend)(0, oldset,
 			newset, regs->esi, regs->edi,
 			regs->ebp, regs->eax,
-			regs->ds, regs->es,
-			regs->fs, regs->gs, regs->orig_eax,
-			regs->eip, regs->cs, regs->eflags,
-			regs->esp, regs->ss);
+			regs->xds, regs->xes,
+			regs->orig_eax,
+			regs->eip, regs->xcs, regs->eflags,
+			regs->esp, regs->xss);
 }
 #endif /* __sparc__ */
