@@ -3,7 +3,7 @@
  *
  *  Copyright 1994-1996  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: socksys.c,v 1.52 1997/11/12 22:30:09 jaggy Exp $
+ * $Id: socksys.c,v 1.67 1998/11/05 21:50:29 jaggy Exp $
  * $Source: /u/CVS/ibcs/iBCSemul/socksys.c,v $
  */
 
@@ -12,10 +12,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/config.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/types.h>
 #include <linux/errno.h>
@@ -28,6 +25,7 @@
 #include <linux/sched.h>
 #include <linux/signal.h>
 #include <linux/socket.h>
+#include <net/sock.h>
 #include <linux/malloc.h>
 #include <linux/mm.h>
 #include <linux/un.h>
@@ -35,6 +33,10 @@
 #include <linux/time.h>
 #include <linux/termios.h>
 #include <linux/sys.h>
+#include <linux/file.h>
+#include <linux/poll.h>
+#include <linux/smp_lock.h>
+#include <linux/capability.h>
 
 #include <ibcs/ibcs.h>
 #include <ibcs/map.h>
@@ -44,38 +46,42 @@
 #endif
 
 #include <ibcs/socksys.h>
-#ifdef EMU_XTI
 #include <ibcs/tli.h>
-#endif
 
 
 int socksys_major;
 
 
-static int socksys_read(struct inode *ino, struct file *filep,
-			char *buf, int count);
-static int socksys_write(struct inode *ino, struct file *filep,
-			const char *buf, int count);
+static int socksys_read(struct file *filep,
+			char *buf, size_t count, loff_t *offset);
+static int socksys_write(struct file *filep,
+			const char *buf, size_t count, loff_t *offset);
 static int socksys_open(struct inode *ino, struct file *filep);
-static void socksys_close(struct inode *ino, struct file *filep);
-static int socksys_select(struct inode *ino, struct file *filep,
-			int sel_type, select_table *wait);
+static int socksys_close(struct inode *ino, struct file *filep);
+static unsigned int socksys_poll(struct file *filep, struct poll_table_struct *wait);
 
 
 /* socksys_fops defines the file operations that can be applied to the
  * /dev/socksys device.
  */
 static struct file_operations socksys_fops = {
-	NULL,			/* lseek */
+	NULL,			/* llseek */
 	socksys_read,		/* read */
 	socksys_write,		/* write */
 	NULL,			/* readdir */
-	NULL,			/* select */
+	NULL,			/* poll */
 	NULL,			/* ioctl - ibcs_ioctl calls ibcs_ioctl_socksys*/
 	NULL,			/* mmap */
 	socksys_open,		/* open */
-	socksys_close,		/* close */
-	NULL			/* fsync */
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,119)
+	NULL,			/* flush */
+#endif
+	socksys_close,		/* release */
+	NULL,			/* fsync */
+	NULL,			/* fasync */
+	NULL,			/* check_media_change */
+	NULL,			/* revalidate */
+	NULL			/* lock */
 };
 
 /* socksys_socket_fops defines the file operations that can be applied to
@@ -83,25 +89,58 @@ static struct file_operations socksys_fops = {
  * created.
  */
 static struct file_operations socksys_socket_fops = {
-	NULL,			/* lseek */
+	NULL,			/* llseek */
 	NULL,			/* read */
 	NULL,			/* write */
 	NULL,			/* readdir */
-	NULL,			/* select */
+	NULL,			/* poll */
 	NULL,			/* ioctl */
 	NULL,			/* mmap */
 	NULL,			/* open */
-	NULL,			/* close */
-	NULL			/* fsync */
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,119)
+	NULL,			/* flush */
+#endif
+	NULL,			/* release */
+	NULL,			/* fsync */
+	NULL,			/* fasync */
+	NULL,			/* check_media_change */
+	NULL,			/* revalidate */
+	NULL			/* lock */
 };
-static void (*sock_close)(struct inode *inode, struct file *file);
-static int (*sock_select)(struct inode *ino, struct file *filep,
-			int sel_type, select_table *wait);
+static int (*sock_close)(struct inode *inode, struct file *file);
+static unsigned int (*sock_poll)(struct file *filep, struct poll_table_struct *wait);
 
 
-static void
-inherit_socksys_funcs(struct file *filep, int fd)
+void
+inherit_socksys_funcs(unsigned int fd, int state)
 {
+#ifdef EMU_XTI
+	struct T_private *priv;
+#endif
+	struct inode *ino;
+	struct file *filep = fget(fd);
+	if (!filep)
+		return;
+
+	ino = filep->f_dentry->d_inode;
+
+#ifdef SO_BSDCOMPAT
+	/* SYSV sockets are BSD like with respect to ICMP errors
+	 * with UDP rather than RFC conforming. I think.
+	 */
+	ino->u.socket_i.sk->bsdism = 1;
+#endif
+
+#ifdef EMU_XTI
+	priv = (struct T_private *)kmalloc(sizeof(struct T_private), GFP_KERNEL);
+	if (priv) {
+		priv->state = state;
+		priv->offset = 0;
+		priv->pfirst = priv->plast = NULL;
+	}
+	filep->private_data = priv;
+#endif
+
 	/* If our file operations don't appear to match
 	 * what the socket system is advertising it may
 	 * be because we haven't initialised ours at all
@@ -111,32 +150,38 @@ inherit_socksys_funcs(struct file *filep, int fd)
 	 * sockets which *should* use the old routines
 	 * until they close - tough, for now.
 	 */
-	if (socksys_socket_fops.read != current->FD[fd]->f_op->read) {
+	if (socksys_socket_fops.read != filep->f_op->read) {
 		memcpy(&socksys_socket_fops,
-			current->FD[fd]->f_op,
+			filep->f_op,
 			sizeof(struct file_operations));
 		sock_close = socksys_socket_fops.release;
-		sock_select = socksys_socket_fops.select;
+		sock_poll = socksys_socket_fops.poll;
 		socksys_socket_fops.release = socksys_close;
-		socksys_socket_fops.select = socksys_select;
+		socksys_socket_fops.poll = socksys_poll;
 	}
 	filep->f_op = &socksys_socket_fops;
+#if 1
+	ino->i_mode = 0020000; /* S_IFCHR */
+	ino->i_rdev = MKDEV(socksys_major, 0);
+#endif
+	fput(filep);
+	MOD_INC_USE_COUNT;
 }
 
 
 static int
-spx_connect(struct inode *ino, struct file *filep, int spxnum)
+spx_connect(unsigned int fd, int spxnum)
 {
 	int newfd, err, args[3];
 	struct sockaddr_un addr = {
 		AF_UNIX, "/tmp/.X11-unix/X0"
 	};
-	int old_fs = get_fs();
+	mm_segment_t old_fs = get_fs();
 
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] %lx choose service %d\n",
-			current->pid, (unsigned long)filep, spxnum);
+		printk(KERN_DEBUG "%d iBCS: SPX: %u choose service %d\n",
+			current->pid, fd, spxnum);
 #endif
 
 	/* Rather than use an explicit path to the X :0 server
@@ -152,7 +197,7 @@ spx_connect(struct inode *ino, struct file *filep, int spxnum)
 	 *   Hey, this is what we do here now!
 	 */
 	addr.sun_family = AF_UNIX;
-	sprintf(addr.sun_path, "/dev/spx.map/%d", spxnum);
+	sprintf(addr.sun_path, "/dev/spx.map/%u", spxnum);
 	set_fs(get_ds());
 	err = SYS(readlink)(addr.sun_path, addr.sun_path,
 				sizeof(addr.sun_path)-1);
@@ -161,9 +206,8 @@ spx_connect(struct inode *ino, struct file *filep, int spxnum)
 #ifdef IBCS_TRACE
 		if (ibcs_trace & TRACE_SOCKSYS)
 			printk(KERN_DEBUG
-				"SPX: [%d] %lx no symlink \"%s\", try X :0\n",
-				current->pid, (unsigned long)filep,
-				addr.sun_path);
+				"%d iBCS: SPX: %u no symlink \"%s\", try X :0\n",
+				current->pid, fd, addr.sun_path);
 #endif
 		strcpy(addr.sun_path, "/tmp/.X11-unix/X0");
 	} else {
@@ -174,8 +218,8 @@ spx_connect(struct inode *ino, struct file *filep, int spxnum)
 
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] %lx get a Unix domain socket\n",
-			current->pid, (unsigned long)filep);
+		printk(KERN_DEBUG "%d iBCS: SPX: %u get a Unix domain socket\n",
+			current->pid, fd);
 #endif
 	args[0] = AF_UNIX;
 	args[1] = SOCK_STREAM;
@@ -188,8 +232,8 @@ spx_connect(struct inode *ino, struct file *filep, int spxnum)
 
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] %lx connect to \"%s\"\n",
-			current->pid, (unsigned long)filep, addr.sun_path);
+		printk(KERN_DEBUG "%d iBCS: SPX: %u connect to \"%s\"\n",
+			current->pid, fd, addr.sun_path);
 #endif
 	args[0] = newfd;
 	args[1] = (int)&addr;
@@ -206,24 +250,34 @@ spx_connect(struct inode *ino, struct file *filep, int spxnum)
 
 
 int
-ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count)
+ibcs_socksys_fd_init(int fd, int rw, const char *buf, int *count)
 {
-	struct file *filep = current->FD[fd];
-	struct inode *ino = current->FD[fd]->f_inode;
+	struct file *filep;
+	struct inode *ino;
 	int error, sockfd;
 
+	/* FIXME: we are always holding the filep when we get here
+	 * anyway so we could just use fcheck here...
+	 */
+	filep = fget(fd);
+	if (!filep)
+		return -EBADF;
+	ino = filep->f_dentry->d_inode;
+
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
 		printk(KERN_DEBUG
-			"socksys: [%d] fd=%d initializing\n",
+			"%d iBCS: socksys: fd=%d initializing\n",
 			current->pid, fd);
 #endif
 	/* Minor = 0 is the socksys device itself. No special handling
 	 *           will be needed as it is controlled by the application
 	 *           via ioctls.
 	 */
-	if (MINOR(ino->i_rdev) == 0)
+	if (MINOR(ino->i_rdev) == 0) {
+		fput(filep);
 		return 0;
+	}
 
 	/* Minor = 1 is the spx device. This is the client side of a
 	 *           streams pipe to the X server. Under SCO and friends
@@ -237,22 +291,26 @@ ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count)
 	if (MINOR(ino->i_rdev) == 1) {
 		int unit = 1;
 
+		fput(filep);
+
 		/* It seems early spx implementations were just a
 		 * quick hack to get X to work. They only supported
 		 * one destination and connected automatically.
 		 * Later versions take a single byte write, the
 		 * value of the byte telling them which destination
-		 * to connect to.
+		 * to connect to. Hence this quick hack to work
+		 * with both. If the first write is a single byte
+		 * it's a connect request otherwise we auto-connect
+		 * to destination 1.
 		 */
-		if (rw == 1 && count == 1) {
-			error = verify_area(VERIFY_READ, buf, 1*sizeof(char));
+		if (rw == 1 && *count == 1) {
+			error = get_user(unit, buf);
 			if (error)
 				return error;
-
-			unit = get_fs_byte(buf);
+			(*count)--;
 		}
 
-		sockfd = spx_connect(ino, filep, unit);
+		sockfd = spx_connect(fd, unit);
 		if (sockfd < 0)
 			return sockfd;
 	}
@@ -266,17 +324,14 @@ ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count)
 	 * type is inferrred from the protocol.
 	 */
 	else { /* XTI */
-#ifdef EMU_XTI
-		struct T_private *priv;
-#endif
 		int args[3];
-		int old_fs = get_fs();
+		mm_segment_t old_fs = get_fs();
 
 		/* Grab a socket. */
 #ifdef IBCS_TRACE
 		if (ibcs_trace & TRACE_SOCKSYS)
 			printk(KERN_DEBUG
-				"XTI: [%d] %d get socket for transport end"
+				"%d iBCS: XTI: %d get socket for transport end"
 				" point (dev = 0x%04x)\n",
 				current->pid, fd, ino->i_rdev);
 #endif
@@ -312,10 +367,11 @@ ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count)
 				args[2] = 0;
 				break;
 		}
+		fput(filep);
 #ifdef IBCS_TRACE
 		if (ibcs_trace & TRACE_SOCKSYS)
 			printk(KERN_DEBUG
-				"XTI: [%d] %d socket %d %d %d\n",
+				"%d iBCS: XTI: %d socket %d %d %d\n",
 				current->pid, fd,
 				args[0], args[1], args[2]);
 #endif
@@ -324,90 +380,92 @@ ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count)
 		set_fs(old_fs);
 		if (sockfd < 0)
 			return sockfd;
-
-#ifdef EMU_XTI
-		priv = (struct T_private *)kmalloc(sizeof(struct T_private), GFP_KERNEL);
-		if (!priv) {
-			SYS(close)(sockfd);
-			return -ENOMEM;
-		}
-		priv->state = TS_UNBND;
-		priv->offset = 0;
-		priv->pfirst = priv->plast = NULL;
-		current->FD[sockfd]->private_data = priv;
-#endif
 	}
 
 	/* Redirect operations on the socket fd via our emulation
 	 * handlers then swap the socket fd and the original fd,
 	 * discarding the original fd.
 	 */
-	inherit_socksys_funcs(current->FD[sockfd], sockfd);
+	inherit_socksys_funcs(sockfd, TS_UNBND);
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "XTI: [%d] %d (ino=%lx) -> %d (ino=%lx)\n",
-			current->pid,
-			fd, (unsigned long)current->FD[fd]->f_inode,
-			sockfd, (unsigned long)current->FD[sockfd]->f_inode);
+		printk(KERN_DEBUG "%d iBCS: XTI: %d -> %d\n",
+			current->pid, fd, sockfd);
 #endif
-#if 0
-	current->FD[sockfd]->f_inode->i_rdev = ino->i_rdev;
-#endif
-	MOD_INC_USE_COUNT;
 	SYS(dup2)(sockfd, fd);
 	SYS(close)(sockfd);
-	return 0;
+	return 1;
 }
 
 
 static int
-socksys_read(struct inode *ino, struct file *filep, char *buf, int count)
+socksys_read(struct file *filep, char *buf, size_t count, loff_t *offset)
 {
+	struct inode *ino = filep->f_dentry->d_inode;
+
+	/* FIXME: this condition *must* be true - if we have already
+	 * replaced this with a socket we have changed the file ops
+	 * too surely?
+	 */
 	if (ino && !ino->i_sock) {
 		int fd;
 		for (fd=0; fd<NR_OPEN; fd++) {
-			if (current->FD[fd] == filep) {
+			if (fcheck(fd) == filep) {
 				int error;
-				error = ibcs_socksys_fd_init(fd, 0, NULL, 0);
-				if (error)
-					return error;
-				filep = current->FD[fd];
-				ino = filep->f_inode;
-				break;
+				error = ibcs_socksys_fd_init(fd, 0, NULL, NULL);
+				if (error > 0) {
+					fput(filep);
+					filep = fget(fd);
+					return filep->f_op->read(filep, buf,
+								count, offset);
+				} else
+					return error ? error : -EINVAL;
 			}
 		}
 	}
-	return filep->f_op->read(ino, filep, buf, count);
+	return -EINVAL;
 }
 
 
 static int
-socksys_write(struct inode *ino, struct file *filep, const char *buf, int count)
+socksys_write(struct file *filep, const char *buf, size_t count, loff_t *offset)
 {
+	struct inode *ino = filep->f_dentry->d_inode;
+
+	/* FIXME: this condition *must* be true - if we have already
+	 * replaced this with a socket we have changed the file ops
+	 * too surely?
+	 */
 	if (ino && !ino->i_sock) {
 		int fd;
 		for (fd=0; fd<NR_OPEN; fd++) {
-			if (current->FD[fd] == filep) {
+			if (fcheck(fd) == filep) {
 				int error;
-				error = ibcs_socksys_fd_init(fd, 1, buf, count);
-				if (error)
-					return error;
-				filep = current->FD[fd];
-				ino = filep->f_inode;
-				break;
+				error = ibcs_socksys_fd_init(fd, 1, buf, &count);
+				if (error > 0) {
+					fput(filep);
+					filep = fget(fd);
+					return count
+						? filep->f_op->write(filep, buf,
+								count, offset)
+						: 0;
+				} else
+					return error ? error : -EINVAL;
 			}
 		}
 	}
-	return filep->f_op->write(ino, filep, buf, count);
+	return -EINVAL;
 }
 
 
 int
 socksys_syscall(int *sp)
 {
-	int cmd;
+	int error, cmd;
 
-	cmd = (int)get_fs_long(sp);
+	error = get_user(cmd, sp);
+	if (error)
+		return error;
 	sp++;
 
 #ifdef IBCS_TRACE
@@ -423,13 +481,14 @@ socksys_syscall(int *sp)
 			"recvmsg", "sendmsg", "sockpair"
 		};
 
-		a0 = (unsigned long)get_fs_long(sp+0);
-		a1 = (unsigned long)get_fs_long(sp+1);
-		a2 = (unsigned long)get_fs_long(sp+2);
-		a3 = (unsigned long)get_fs_long(sp+3);
-		a4 = (unsigned long)get_fs_long(sp+4);
-		a5 = (unsigned long)get_fs_long(sp+5);
-		printk(KERN_DEBUG "socksys: [%d]: %s (%d) <0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx>\n",
+		get_user(a0, sp+0);
+		get_user(a1, sp+1);
+		get_user(a2, sp+2);
+		get_user(a3, sp+3);
+		get_user(a4, sp+4);
+		get_user(a5, sp+5);
+		printk(KERN_DEBUG "%d iBCS: socksys:"
+			" %s (%d) <0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx>\n",
 			current->pid,
 			(cmd >= 0 && cmd < sizeof(cmd_map)/sizeof(cmd_map[0]))
 				? cmd_map[cmd] : "???",
@@ -444,23 +503,19 @@ socksys_syscall(int *sp)
 			 * right thing for ioctls.
 			 */
 			int fd;
+			unsigned long x;
 
-			put_fs_long(
-				map_value(af_map,
-					get_fs_long(((unsigned long *)sp)+0),
-					0),
+			get_user(x, ((unsigned long *)sp)+0);
+			put_user(map_value(af_map, x, 0),
  				((unsigned long *)sp)+0);
-			put_fs_long(
-				map_value(type_map,
-					get_fs_long(((unsigned long *)sp)+1),
-					0),
+			get_user(x, ((unsigned long *)sp)+1);
+			put_user(map_value(type_map, x, 0),
  				((unsigned long *)sp)+1);
 
 			if ((fd = SYS(socketcall)(SYS_SOCKET, sp)) < 0)
 				return fd;
 
-			inherit_socksys_funcs(current->FD[fd], fd);
-			MOD_INC_USE_COUNT;
+			inherit_socksys_funcs(fd, TS_UNBND);
 			return fd;
 		}
 
@@ -470,36 +525,13 @@ socksys_syscall(int *sp)
 			if ((fd = SYS(socketcall)(SYS_ACCEPT, sp)) < 0)
 				return fd;
 
-			inherit_socksys_funcs(current->FD[fd], fd);
-			MOD_INC_USE_COUNT;
+			inherit_socksys_funcs(fd, TS_DATA_XFER);
 			return fd;
 		}
 		case SSYS_SO_BIND:
 			return SYS(socketcall)(SYS_BIND, sp);
-		case SSYS_SO_CONNECT: {
-#ifndef INIT_MM
-			struct sockaddr *sa;
-
-			/* Trap attack... Some things (notably SVR
-			 * syslog code) seems to think 0.0.0.0 is
-			 * a valid address for localhost.
-			 */
-			sa = (struct sockaddr *)get_fs_long(sp+1);
-			if (get_fs_word(&sa->sa_family) == AF_INET) {
-				unsigned long *addr;
-
-				addr = &(((struct sockaddr_in *)sa)->sin_addr.s_addr);
-				if (get_fs_long(addr) == INADDR_ANY) {
-					put_fs_long(0x0100007f, addr);
-#ifdef IBCS_TRACE
-					if (ibcs_trace & TRACE_SOCKSYS)
-						printk(KERN_DEBUG "iBCS: socksys: remapped INADDR_ANY to localhost\n");
-#endif
-				}
-			}
-#endif
+		case SSYS_SO_CONNECT:
 			return SYS(socketcall)(SYS_CONNECT, sp);
-		}
 		case SSYS_SO_GETPEERNAME:
 			return SYS(socketcall)(SYS_GETPEERNAME, sp);
 		case SSYS_SO_GETSOCKNAME:
@@ -537,52 +569,61 @@ socksys_syscall(int *sp)
 			int error, len;
 			char *name, *p;
 
-			if ((error = verify_area(VERIFY_READ, sp, 2*sizeof(long))))
-				return error;
-
-			name = (char *)get_fs_long(sp+0);
-			len = (int)get_fs_long(sp+1);
-			if ((error = verify_area(VERIFY_WRITE, name, len)))
-				return error;
-
-			--len;
-			for (p=system_utsname.nodename; *p && *p != '.'; p++);
-			if (*p == '.')
-				p++;
-			else
-				p = system_utsname.domainname;
-			if (strcmp(p, "(none)"))
-				for (;*p && len > 0; p++,len--)
-					put_fs_byte(*p, name++);
-			put_fs_byte('\0', name);
-
-			return (0);
+			error = get_user(name, (char *)(sp+0));
+			if (!error)
+				get_user(len, sp+1);
+			if (!error) {
+				down(&uts_sem);
+				error = verify_area(VERIFY_WRITE, name, len);
+				if (!error) {
+					--len;
+					for (p=system_utsname.nodename; *p && *p != '.'; p++);
+					if (*p == '.')
+						p++;
+					else
+						p = system_utsname.domainname;
+					if (strcmp(p, "(none)"))
+						for (;*p && len > 0; p++,len--) {
+							__put_user(*p, name);
+							name++;
+						}
+					__put_user('\0', name);
+				}
+				up(&uts_sem);
+			}
+			return error;
 		}
 		case SSYS_SO_SETIPDOMAIN: {
 			int error, len, togo;
 			char *name, *p;
 
-			if (!suser())
+			if (!capable(CAP_SYS_ADMIN))
 				return -EPERM;
 
-			if ((error = verify_area(VERIFY_READ, sp, 2*sizeof(long))))
+			error = get_user(name, (char *)(sp+0));
+			if (!error)
+				error = get_user(len, sp+1);
+			if (error)
 				return error;
 
-			name = (char *)get_fs_long(sp+0);
-			len = (int)get_fs_long(sp+1);
-
+			down(&uts_sem);
 			togo = __NEW_UTS_LEN;
 			for (p=system_utsname.nodename; *p && *p != '.'; p++,togo--);
 			if (*p == '.')
 				p++,togo--;
 
-			if (len > togo)
-				return -EINVAL;
-
-			while (len-- > 0)
-				*p++ = get_fs_byte(name++);
-			*p = '\0';
-			return 0;
+			error = -EINVAL;
+			if (len <= togo) {
+				while (len-- > 0) {
+					get_user(*p, name);
+					p++;
+					name++;
+				}
+				*p = '\0';
+				error = 0;
+			}
+			up(&uts_sem);
+			return error;
 		}
 
 		case SSYS_SO_SETREUID:
@@ -590,10 +631,11 @@ socksys_syscall(int *sp)
 			int error;
 			uid_t ruid, euid;
 
-			if ((error = verify_area(VERIFY_READ, sp, 2*sizeof(long))))
+			error = get_user(ruid, sp+0);
+			if (!error)
+				error = get_user(euid, sp+1);
+			if (error)
 				return error;
-			ruid = (uid_t)get_fs_long(sp+0);
-			euid = (uid_t)get_fs_long(sp+1);
 			return (cmd == SSYS_SO_SETREUID)
 				? SYS(setreuid)(ruid, euid)
 				: SYS(setregid)(ruid, euid);
@@ -605,10 +647,11 @@ socksys_syscall(int *sp)
 			struct timeval *tv;
 			struct timezone *tz;
 
-			if ((error = verify_area(VERIFY_READ, sp, 2*sizeof(long))))
+			error = get_user(tv, sp+0);
+			if (!error)
+				error = get_user(tz, sp+1);
+			if (error)
 				return error;
-			tv = (struct timeval *)get_fs_long(sp+0);
-			tz = (struct timezone *)get_fs_long(sp+1);
 			return (cmd == SSYS_SO_GETTIME)
 				? SYS(gettimeofday)(tv, tz)
 				: SYS(settimeofday)(tv, tz);
@@ -618,21 +661,24 @@ socksys_syscall(int *sp)
 			int error, which;
 			struct itimerval *value;
 
-			if ((error = verify_area(VERIFY_READ, sp, 2*sizeof(long))))
+			error = get_user(which, sp+0);
+			if (!error)
+				error = get_user(value, sp+1);
+			if (error)
 				return error;
-			which = (int)get_fs_long(sp+0);
-			value = (struct itimerval *)get_fs_long(sp+1);
 			return SYS(getitimer)(which, value);
 		}
 		case SSYS_SO_SETITIMER: {
 			int error, which;
 			struct itimerval *value, *ovalue;
 
-			if ((error = verify_area(VERIFY_READ, sp, 3*sizeof(long))))
+			error = get_user(which, sp+0);
+			if (!error)
+				error = get_user(value, sp+1);
+			if (!error)
+				error = get_user(ovalue, sp+2);
+			if (error)
 				return error;
-			which = (int)get_fs_long(sp+0);
-			value = (struct itimerval *)get_fs_long(sp+1);
-			ovalue = (struct itimerval *)get_fs_long(sp+2);
 			return SYS(setitimer)(which, value, ovalue);
 		}
 
@@ -665,32 +711,43 @@ socksys_syscall(int *sp)
 			 * operations with our own so we can do the
 			 * right thing for ioctls.
 			 */
+			struct file *filep;
 			struct inode *ino;
 			int error, args[4], pairin[2], pairout[2];
-			int old_fs = get_fs();
+			mm_segment_t old_fs = get_fs();
 
 			/* The first two arguments are file descriptors
 			 * of sockets which have already been opened
 			 * and should now be connected back to back.
 			 */
-			if ((error = verify_area(VERIFY_READ, sp, 2*sizeof(long))))
+			error = get_user(pairin[0], sp+0);
+			if (!error)
+				error = get_user(pairin[1], sp+1);
+			if (error)
 				return error;
-			pairin[0] = (int)get_fs_long(sp+0);
-			pairin[1] = (int)get_fs_long(sp+1);
 
-			if (pairin[0] >= NR_OPEN
-			|| !current->FD[pairin[0]]
-			|| !(ino = current->FD[pairin[0]]->f_inode)
-			|| !ino->i_sock)
+			filep = fget(pairin[0]);
+			if (!filep)
 				return -EBADF;
+			ino = filep->f_dentry->d_inode;
+			if (!ino || !ino->i_sock) {
+				fput(filep);
+				return -EBADF;
+			}
 
 			args[0] = AF_UNIX;
 			args[1] = ino->u.socket_i.type;
 			args[2] = 0;
 			args[3] = (int)pairout;
 
+			fput(filep);
+
+			/* FIXME: Do we need to close these here? If we
+			 * fail to connect them should they be open?
+			 */
 			SYS(close)(pairin[0]);
 			SYS(close)(pairin[1]);
+
 			set_fs(get_ds());
 			error = SYS(socketcall)(SYS_SOCKETPAIR, args);
 			set_fs(old_fs);
@@ -706,10 +763,8 @@ socksys_syscall(int *sp)
 				SYS(close)(pairout[1]);
 			}
 
-			inherit_socksys_funcs(current->FD[pairin[0]], pairin[0]);
-			inherit_socksys_funcs(current->FD[pairin[1]], pairin[1]);
-			MOD_INC_USE_COUNT;
-			MOD_INC_USE_COUNT;
+			inherit_socksys_funcs(pairin[0], TS_DATA_XFER);
+			inherit_socksys_funcs(pairin[1], TS_DATA_XFER);
 			return 0;
 		}
 	}
@@ -731,26 +786,31 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 			struct domnam_args dn;
 			char *p;
 
-			if ((error = verify_area(VERIFY_READ, (char *)arg, sizeof(struct domnam_args))))
-				return error;
-			memcpy_fromfs(&dn, (char *)arg, sizeof(struct domnam_args));
+			error = copy_from_user(&dn, (char *)arg,
+					sizeof(struct domnam_args));
+			if (error)
+				return -EFAULT;
 
-			if ((error = verify_area(VERIFY_WRITE, dn.name, dn.namelen)))
-				return error;
-
-			--dn.namelen;
-			for (p=system_utsname.domainname; *p && dn.namelen > 0; p++,dn.namelen--)
-				put_fs_byte(*p, dn.name++);
-			put_fs_byte('\0', dn.name);
-
-			return (0);
+			down(&uts_sem);
+			error = verify_area(VERIFY_WRITE, dn.name, dn.namelen);
+			if (!error) {
+				--dn.namelen;
+				for (p=system_utsname.domainname; *p && dn.namelen > 0; p++,dn.namelen--) {
+					__put_user(*p, dn.name);
+					dn.name++;
+				}
+				__put_user('\0', dn.name);
+			}
+			up(&uts_sem);
+			return error;
 		}
 		case NIOCSETDOMNAM: {
 			struct domnam_args dn;
 
-			if ((error = verify_area(VERIFY_READ, (char *)arg, sizeof(struct domnam_args))))
-				return error;
-			memcpy_fromfs(&dn, (char *)arg, sizeof(struct domnam_args));
+			error = copy_from_user(&dn, (char *)arg,
+					sizeof(struct domnam_args));
+			if (error)
+				return -EFAULT;
 
 			return SYS(setdomainname)(dn.name, dn.namelen);
 		}
@@ -761,9 +821,10 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 			 */
 			struct lstat_args st;
 
-			if ((error = verify_area(VERIFY_READ, (char *)arg, sizeof(struct lstat_args))))
-				return error;
-			memcpy_fromfs(&st, (char *)arg, sizeof(struct lstat_args));
+			error = copy_from_user(&st, (char *)arg,
+					sizeof(struct lstat_args));
+			if (error)
+				return -EFAULT;
 
 			return ibcs_lstat(st.fname, st.statb);
 		}
@@ -771,28 +832,35 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 		case NIOCOLDGETFH:
 		case NIOCGETFH: {
 			struct getfh_args gf;
-			struct inode *ino;
+			struct dentry *dentry;
 
-			if (!suser())
+			if (!capable(CAP_SYS_ADMIN))
 				return -EPERM;
 
-			if ((error = verify_area(VERIFY_READ, (char *)arg, sizeof(struct getfh_args))))
-				return error;
-			memcpy_fromfs(&gf, (char *)arg, sizeof(struct getfh_args));
+			error = copy_from_user(&gf, (char *)arg,
+						sizeof(struct getfh_args));
+			if (error)
+				return -EFAULT;
 
-			if ((error = verify_area(VERIFY_WRITE, (char *)gf.fhp, sizeof(fhandle_t))))
-				return error;
-
-			if ((error = namei(gf.fname, &ino)))
-				return error;
-
-			put_fs_word(ino->i_dev, &gf.fhp->fh.fsid);
-			put_fs_long(ino->i_ino, &gf.fhp->fh.fno);
-			put_fs_long(0L, &gf.fhp->fh.fgen);
-			put_fs_word(ino->i_dev, &gf.fhp->fh.ex_fsid);
-			put_fs_long(ino->i_ino, &gf.fhp->fh.ex_fno);
-			put_fs_long(0L, &gf.fhp->fh.ex_fgen);
-			return 0;
+			lock_kernel();
+			error = verify_area(VERIFY_WRITE, (char *)gf.fhp, sizeof(fhandle_t));
+			if (!error) {
+				dentry = namei(gf.fname);
+				error = PTR_ERR(dentry);
+				if (!IS_ERR(dentry)) {
+					struct inode *ino = dentry->d_inode;
+					__put_user(ino->i_dev, &gf.fhp->fh.fsid);
+					__put_user(ino->i_ino, &gf.fhp->fh.fno);
+					__put_user(0L, &gf.fhp->fh.fgen);
+					__put_user(ino->i_dev, &gf.fhp->fh.ex_fsid);
+					__put_user(ino->i_ino, &gf.fhp->fh.ex_fno);
+					__put_user(0L, &gf.fhp->fh.ex_fgen);
+					dput(dentry);
+					error = 0;
+				}
+			}
+			unlock_kernel();
+			return error;
 		}
 
 		case NIOCNFSD:
@@ -811,6 +879,7 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 		case SVR4_SIOCSLOWAT:
 			/* Linux doesn't support them but lie anyway
 			 * or some things take it as fatal (why?)
+			 * FIXME: actually we can do this now...
 			 */
 			return 0;
 
@@ -820,11 +889,12 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 		case SVR4_SIOCGLOWAT:
 			/* Linux doesn't support them but lie anyway
 			 * or some things take it as fatal (why?)
+			 * FIXME: actually we can do this now...
 			 */
 			if ((error = verify_area(VERIFY_WRITE, (char *)arg,
 						sizeof(unsigned long))))
 				return error;
-			put_fs_long(0, (unsigned long *)arg);
+			put_user(0, (unsigned long *)arg);
 			return 0;
 
 		case SSYS_SIOCATMARK:		/* at oob mark? */
@@ -842,10 +912,13 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 		case SSYS_FIONREAD:		/* BSD compatibilty */
 			error = SYS(ioctl)(fd, TIOCINQ, arg);
 #ifdef IBCS_TRACE
-			if (!error && (ibcs_trace & TRACE_SOCKSYS))
+			if (!error && (ibcs_trace & TRACE_SOCKSYS)) {
+				unsigned long n;
+				get_user(n, (unsigned long *)arg);
 				printk(KERN_DEBUG
-					"iBCS: socksys FIONREAD found %lu bytes ready\n",
-					(unsigned long)get_fs_long(arg));
+					"%d iBCS: socksys: %d FIONREAD found %lu bytes ready\n",
+					current->pid, fd, n);
+			}
 #endif
 			return error;
 
@@ -941,7 +1014,8 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 		{
 			struct sockaddr uaddr;
 			int uaddr_len = sizeof(struct sockaddr);
-			int op, args[3], old_fs;
+			int op, args[3];
+			mm_segment_t old_fs;
 
 			if ((error = verify_area(VERIFY_WRITE, (char *)arg, sizeof(struct sockaddr))))
 				return error;
@@ -957,7 +1031,7 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 			error = SYS(socketcall)(op, args);
 			set_fs(old_fs);
 			if (error >= 0)
-				memcpy_tofs((char *)arg, &uaddr, uaddr_len);
+				copy_to_user((char *)arg, &uaddr, uaddr_len);
 			return error;
 		}
 
@@ -986,7 +1060,7 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 		case SVR4_SIOCX25TBL:
 
 		default:
-			printk(KERN_DEBUG "socksys: [%d] %d: ioctl 0x%x with argument 0x%lx requested\n",
+			printk(KERN_DEBUG "%d iBCS: socksys: %d: ioctl 0x%x with argument 0x%lx requested\n",
 				current->pid, fd,
 				cmd, (unsigned long)arg);
 			break;
@@ -1003,38 +1077,46 @@ socksys_open(struct inode *ino, struct file *filep)
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
 		printk(KERN_DEBUG
-			"socksys: [%d] filep=0x%08lx, inode=0x%08lx opening\n",
+			"%d iBCS: socksys: filep=0x%08lx, inode=0x%08lx opening\n",
 			current->pid, (unsigned long)filep, (unsigned long)ino);
 #endif
 	return 0;
 }
 
 
-static int
-socksys_select(struct inode *ino, struct file *filep,
-		int sel_type, select_table *wait)
+static unsigned int
+socksys_poll(struct file *filep, struct poll_table_struct *wait)
 {
+	unsigned int mask = 0;
+
 #ifdef EMU_XTI
-	/* If we are checking for readable data and this is a
-	 * timod transport end point and there is a control
-	 * message queued we have readable data.
+	struct inode *ino = filep->f_dentry->d_inode;
+
+	/* If this is a timod transport end point and there
+	 * is a control message queued we have readable data.
 	 */
-	if (sel_type == SEL_IN
-	&& ino && ino->i_sock && MINOR(ino->i_rdev) != 1
-	&& filep->private_data
-	&& ((struct T_private *)filep->private_data)->pfirst)
-		return 1;
+	if (ino && ino->i_sock && MINOR(ino->i_rdev) != 1
+	&& Priv(filep) && Priv(filep)->pfirst)
+		mask = Priv(filep)->pfirst->pri == MSG_HIPRI
+			? POLLPRI
+			: POLLIN;
 #endif
 
-	if (sock_select)
-		return (*sock_select)(ino, filep, sel_type, wait);
-	return 0;
+	if (sock_poll)
+		mask |= (*sock_poll)(filep, wait);
+	return mask;
 }
 
 
-static void
+static int
 socksys_close(struct inode *ino, struct file *filep)
 {
+	int error;
+
+	/* Not being a socket is not an error - it is probably
+	 * just the pseudo device transport provider.
+	 */
+	error = 0;
 	if (ino && ino->i_sock) {
 #ifdef EMU_XTI
 		if (filep->private_data) {
@@ -1048,15 +1130,16 @@ socksys_close(struct inode *ino, struct file *filep)
 			kfree(filep->private_data);
 		}
 #endif
-		sock_close(ino, filep);
+		error = sock_close(ino, filep);
 	}
 
 #ifdef IBCS_TRACE
 	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "socksys: [%d] %lx closed\n",
+		printk(KERN_DEBUG "%d iBCS: socksys: %lx closed\n",
 			current->pid, (unsigned long)filep);
 #endif
 	MOD_DEC_USE_COUNT;
+	return error;
 }
 
 
