@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 1995  Mike Jagdis
  *
- * $Id: svr4.c,v 1.3 1995/11/29 13:05:01 mike Exp $
+ * $Id: svr4.c,v 1.6 1998/06/24 20:34:17 jaggy Exp $
  * $Source: /u/CVS/ibcs/iBCSemul/svr4.c,v $
  */
 
@@ -12,10 +12,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/segment.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/types.h>
 #include <linux/errno.h>
@@ -60,12 +57,8 @@ int svr4_getgroups(int n, unsigned long *buf)
 		if (i)
 			return i;
 	}
-	for (i = 0 ; (i < NGROUPS) && (current->groups[i] != NOGROUP) ; i++) {
-		if (!n)
-			continue;
-		if (i >= n)
-			break;
-		put_fs_long(current->groups[i], buf);
+	for (i = 0; i < current->ngroups && i < n; i++) {
+		put_user(current->groups[i], buf);
 		buf++;
 	}
 	return(i);
@@ -76,23 +69,24 @@ int svr4_setgroups(int n, unsigned long *buf)
 {
 	int i;
 
-	if (!suser())
+	if (!capable(CAP_SETGID))
 		return -EPERM;
 	if (n > NGROUPS)
 		return -EINVAL;
+	/* FIXME: Yuk! What if we hit a bad address? */
 	for (i = 0; i < n; i++, buf++) {
-		current->groups[i] = get_fs_long(buf);
+		get_user(current->groups[i], buf);
 	}
-	if (i < NGROUPS)
-		current->groups[i] = NOGROUP;
+	current->ngroups = n;
 	return 0;
 }
 
 
-int svr4_waitid(int idtype, int id, struct siginfo *infop, int options)
+int svr4_waitid(int idtype, int id, struct svr4_siginfo *infop, int options)
 {
 	long result, kopt;
-	int old_fs, pid, status;
+	mm_segment_t old_fs;
+	int pid, status;
 
 	switch (idtype) {
 		case 0: /* P_PID */
@@ -113,7 +107,7 @@ int svr4_waitid(int idtype, int id, struct siginfo *infop, int options)
 
 	if (infop) {
 		result = verify_area(VERIFY_WRITE, infop,
-					sizeof(struct siginfo));
+					sizeof(struct svr4_siginfo));
 		if (result)
 			return result;
 	}
@@ -132,31 +126,31 @@ int svr4_waitid(int idtype, int id, struct siginfo *infop, int options)
 	if (infop) {
 		unsigned long op, st;
 
-		put_fs_long(current->exec_domain->signal_map[SIGCHLD],
+		put_user(current->exec_domain->signal_map[SIGCHLD],
 			&infop->si_signo);
-		put_fs_long(result,
+		put_user(result,
 			&infop->_data._proc._pid);
 
 		if ((status & 0xff) == 0) {
 			/* Normal exit. */
-			op = CLD_EXITED;
+			op = SVR4_CLD_EXITED;
 			st = status >> 8;
 		} else if ((status & 0xff) == 0x7f) {
 			/* Stopped. */
 			st = (status & 0xff00) >> 8;
 			op = (st == SIGSTOP || st == SIGTSTP)
-				? CLD_STOPPED
-				: CLD_CONTINUED;
+				? SVR4_CLD_STOPPED
+				: SVR4_CLD_CONTINUED;
 			st = current->exec_domain->signal_invmap[st];
 		} else {
 			st = (status & 0xff00) >> 8;
 			op = (status & 0200)
-				? CLD_DUMPED
-				: CLD_KILLED;
+				? SVR4_CLD_DUMPED
+				: SVR4_CLD_KILLED;
 			st = current->exec_domain->signal_invmap[st];
 		}
-		put_fs_long(op, &infop->si_code);
-		put_fs_long(st, &infop->_data._proc._pdata._cld._status);
+		put_user(op, &infop->si_code);
+		put_user(st, &infop->_data._proc._pdata._cld._status);
 	}
 	return 0;
 }
