@@ -7,7 +7,7 @@
  *  April 9 1994, corrected file size passed to/from setrlimit/getrlimit
  *    -- Graham Adams (gadams@ddrive.demon.co.uk)
  *
- * $Id: ulimit.c,v 1.6 1997/03/11 21:30:36 jaggy Exp $
+ * $Id: ulimit.c,v 1.8 1998/06/24 21:18:34 jaggy Exp $
  * $Source: /u/CVS/ibcs/iBCSemul/ulimit.c,v $
  */
 
@@ -16,10 +16,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/segment.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/errno.h>
 #include <linux/sched.h>
@@ -33,6 +30,7 @@
 #include <linux/fs.h>
 #include <linux/sys.h>
 #include <linux/resource.h>
+#include <linux/capability.h>
 
 #include <ibcs/ibcs.h>
 #include <ibcs/trace.h>
@@ -61,7 +59,7 @@ ibcs_ulimit (int cmd, int val)
 				return -ERANGE;
 			val <<= ULIM_BLOCKSIZE_BITS;
 			if (val > current->rlim[RLIMIT_FSIZE].rlim_max) {
-				if (!suser())
+				if (!capable(CAP_SYS_RESOURCE))
 					return -EPERM;
 				else {
 					current->rlim[RLIMIT_FSIZE].rlim_max = val;
