@@ -17,8 +17,8 @@
  * Written by Scott Michel, scottm@intime.com
  * (c) 1994 Scott Michel as part of the Linux iBCS-2 emulator project.
  *
- * $Id: vtkd.c,v 1.5 1995/04/12 10:12:37 mike Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/vtkd.c,v $
+ * $Id: vtkd.c,v 1.6 1998/06/09 19:03:50 jaggy Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/vtkd.c,v $
  */
 
 #include <linux/config.h>
@@ -26,10 +26,7 @@
 #include <linux/module.h>
 #include <linux/version.h>
 
-#include <asm/segment.h>
-#ifndef KERNEL_DS
-#include <linux/segment.h>
-#endif
+#include <asm/uaccess.h>
 
 #include <linux/errno.h>
 #include <linux/fs.h>
