@@ -338,10 +338,11 @@ struct T_private {
 	struct T_primsg *pfirst, *plast;
 };
 
-#define Priv(fd)	((struct T_private *)(current->FD[fd]->private_data))
+#define Priv(file)	((struct T_private *)(file->private_data))
 
 extern int timod_ioctl(struct pt_regs *, int, unsigned int, void *, int, int *);
 extern int timod_putmsg(int, struct inode *, int, struct pt_regs *);
 extern int timod_getmsg(int, struct inode *, int, struct pt_regs *);
+extern int timod_update_socket(int, struct file *, struct pt_regs *);
 
 #endif /* __TLI_H */
